package gs.smolban.db.doobie

import cats.MonadError
import cats.data.EitherT
import cats.syntax.all.*
import doobie.*
import doobie.implicits.*
import gs.smolban.db.DbError
import gs.smolban.db.GroupDb
import gs.smolban.db.doobie.DoobieTypes.*
import gs.smolban.model.CreatedAt
import gs.smolban.model.Group

final class DoobieGroupDb() extends GroupDb[ConnectionIO]:

  import DoobieGroupDb.ErrorCodes
  import DoobieGroupDb.Sql

  /** @inheritdoc
    */
  override def createGroup(
    group: Group
  ): EitherT[ConnectionIO, DbError, Group] =
    EitherT(
      Sql.createGroup(group).run.attemptSql.flatMap {
        case Left(ex) =>
          if ErrorCodes.UniqueConstraintFailed.contains(ex.getErrorCode()) then
            Left(
              DbError.GroupAlreadyExists(group.id, group.slug)
            ).pure[ConnectionIO]
          else MonadError[ConnectionIO, Throwable].raiseError(ex)
        case Right(_) =>
          Right(group).pure[ConnectionIO]
      }
    )

  /** @inheritdoc
    */
  override def readGroup(id: Group.Id): ConnectionIO[Option[Group]] =
    Sql.readGroup(id).option

  /** @inheritdoc
    */
  override def deleteGroup(id: Group.Id): ConnectionIO[Boolean] =
    for
      _    <- Sql.deleteAllTicketsForGroup(id).run
      rows <- Sql.deleteGroup(id).run
    yield rows > 0

object DoobieGroupDb:

  private object Sql:

    def createGroup(group: Group): Update0 = sql"""
    INSERT INTO groups (group_id, slug, created_at)
    VALUES (${group.id}, ${group.slug}, ${group.createdAt})
    """.update

    def readGroup(groupId: Group.Id): Query0[Group] = sql"""
    SELECT slug, created_at
    FROM groups
    WHERE group_id = $groupId
    """.query[Group]

    def deleteAllTicketsForGroup(groupId: Group.Id): Update0 = sql"""
    DELETE FROM tickets
    WHERE group_id = $groupId
    """.update

    def deleteGroup(groupId: Group.Id): Update0 = sql"""
    DELETE FROM groups
    WHERE group_id = $groupId
    """.update

  end Sql

  object ErrorCodes:

    val UniqueConstraintFailed: List[Int] =
      List(Sqlite.UniqueConstraintFailed)

    object Sqlite:

      val UniqueConstraintFailed: Int = 2067

    end Sqlite

  end ErrorCodes

end DoobieGroupDb
