package gs.smolban.model

import cats.Show
import java.time.Instant

/** Tickets represent some tracked work.
  *
  * @param id
  *   Unique identifier _within_ the [[Group]].
  * @param group
  *   Unique identifier of the [[Group]] that owns this ticket.
  * @param createdAt
  *   The instant at which this ticket was created.
  * @param title
  *   Arbitrary string title of the ticket.
  * @param description
  *   Markdown contents of the ticket.
  * @param tags
  *   List of [[Tag]] applied to this ticket.
  * @param status
  *   Current [[Ticket.Status]] of this ticket.
  * @param statusHistory
  *   Linear history of this ticket in terms of status changes.
  */
case class Ticket(
  id: Ticket.Id,
  group: Group.Id,
  createdAt: CreatedAt,
  title: String,
  description: String,
  tags: List[Tag],
  status: Ticket.Status,
  statusHistory: List[(Ticket.Status, Instant)]
):
  lazy val reference: Ticket.Reference = Ticket.Reference(id, group)

object Ticket:

  /** Composite reference that uniquely addresses any [[Ticket]].
    *
    * @param id
    *   The ticket's unique identifier within the [[Group]].
    * @param group
    *   The unique identifier of the [[Group]].
    */
  case class Reference(
    id: Ticket.Id,
    group: Group.Id
  )

  /** Unique identifier - relative to some [[Group]] - for a [[Ticket]]. This is
    * an opaque type for a Long. In general, [[Ticket]] identifiers are
    * sequences within a group.
    */
  opaque type Id = Long

  object Id:

    /** Instantiate a new [[Ticket.Id]].
      *
      * @param id
      *   The underlying Long.
      * @return
      *   The new [[Ticket.Id]] instance.
      */
    def apply(id: Long): Id = id

    given CanEqual[Id, Id] = CanEqual.derived

    given Show[Id] = _.toLong().toString()

    extension (id: Id)
      /** Unwrap this Ticket ID.
        *
        * @return
        *   The underlying Long value.
        */
      def toLong(): Long = id

  end Id

  /** Enumeration that describes the status of a [[Ticket]] in Smolban. Smolban
    * does not yet support custom status/workflow.
    */
  sealed abstract class Status(val name: String)

  object Status:

    given CanEqual[Status, Status] = CanEqual.derived

    given Show[Status] = _.name

    /** This ticket is new, and ready to be started. New tickets may be put into
      * progress or canceled.
      */
    case object Ready extends Status("ready")

    /** This ticket is being worked on actively. In progress tickets can be
      * paused, completed, or canceled.
      */
    case object InProgress extends Status("in_progress")

    /** This ticket was being worked on, but was temporarily stopped. Paused
      * tickets may be put into progress or canceled.
      */
    case object Paused extends Status("paused")

    /** This ticket was driven to completion. The work is done. Once in this
      * state, the status may no longer change.
      */
    case object Complete extends Status("complete")

    /** This ticket was canceled for some reason. Once in this state, the status
      * may no longer change.
      */
    case object Canceled extends Status("canceled")

    val All: List[Status] = List(Ready, InProgress, Paused, Complete, Canceled)

    def parse(candidate: String): Option[Status] =
      All.find(_.name.equalsIgnoreCase(candidate))

  end Status

end Ticket
